"""
This module contains a transformer that applies capping to numeric columns.
"""

import datetime
import warnings
import numpy as np
import pandas as pd

from tubular.base import BaseTransformer


class DateDiffLeapYearTransformer(BaseTransformer):
    """Transformer to calculate the number of years between two dates.

    Parameters
    ----------
    column_lower : str
        Name of date column to subtract.

    column_upper : str
        Name of date column to subtract from.

    new_column_name : str
        Name for the new year column.

    drop_cols : bool
        Flag for whether to drop the original columns.

    missing_replacement : int/float/str
        Value to output if either the lower date value or the upper date value are
        missing. Default value is None.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    column_lower : str
        Name of date column to subtract. This attribute is not for use in any method,
        use 'columns' instead. Here only as a fix to allow string representation of transformer.

    column_upper : str
        Name of date column to subtract from. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : list
        List containing column names for transformation in format [column_lower, column_upper]

    new_column_name : str
        Column name for the year column calculated in the transform method.

    drop_cols : bool
        Indicator whether to drop old columns during transform method.

    """

    def __init__(
        self,
        column_lower,
        column_upper,
        new_column_name,
        drop_cols,
        missing_replacement=None,
        **kwargs,
    ):

        if not isinstance(column_lower, str):
            raise TypeError("column_lower should be a str")

        if not isinstance(column_upper, str):
            raise TypeError("column_upper should be a str")

        if not isinstance(new_column_name, str):
            raise TypeError("new_column_name should be a str")

        if not isinstance(drop_cols, bool):
            raise TypeError("drop_cols should be a bool")

        if missing_replacement:
            if not type(missing_replacement) in [int, float, str]:
                raise TypeError(
                    "if not None, missing_replacement should be an int, float or string"
                )

        super().__init__(columns=[column_lower, column_upper], **kwargs)

        self.new_column_name = new_column_name
        self.drop_cols = drop_cols
        self.missing_replacement = missing_replacement

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper

    def calculate_age(self, row):
        """Function to calculate age from two date columns in a pd.DataFrame.

        This function, although slower than the np.timedelta64 solution (or something
        similar), accounts for leap years to accurately calculate age for all values.

        Parameters
        ----------
        row : pd.Series
            Named pandas series, with lower_date_name and upper_date_name as index values.

        Returns
        -------
        age : int
            Year gap between the upper and lower date values passes.

        """

        if not isinstance(row, pd.Series):
            raise TypeError("row should be a pd.Series")

        if (pd.isnull(row[self.columns[0]])) or (pd.isnull(row[self.columns[1]])):
            return self.missing_replacement

        else:

            if not type(row[self.columns[1]]) in [datetime.date, datetime.datetime]:
                raise TypeError(
                    "upper column values should be datetime.datetime or datetime.date objects"
                )

            if not type(row[self.columns[0]]) in [datetime.date, datetime.datetime]:
                raise TypeError(
                    "lower column values should be datetime.datetime or datetime.date objects"
                )

            age = row[self.columns[1]].year - row[self.columns[0]].year

            if age > 0:
                if (row[self.columns[1]].month, row[self.columns[1]].day) < (
                    row[self.columns[0]].month,
                    row[self.columns[0]].day,
                ):
                    age += -1
            elif age < 0:
                if (row[self.columns[1]].month, row[self.columns[1]].day) > (
                    row[self.columns[0]].month,
                    row[self.columns[0]].day,
                ):
                    age += 1

            return age

    def transform(self, X):
        """Calculate year gap between the two provided columns.

        New column is created under the 'new_column_name', and optionally removes the
        old date columns.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing column_upper and column_lower.

        Returns
        -------
        X : pd.DataFrame
            Transformed data with new_column_name column.

        """

        X = super().transform(X)

        X[self.new_column_name] = X.apply(lambda x: self.calculate_age(x), axis=1)

        if self.drop_cols:
            X.drop(self.columns, axis=1, inplace=True)

        return X


class DateDifferenceTransformer(BaseTransformer):
    """Class to transform calculate the difference between 2 date fields in specified units.

    Parameters
    ----------
    column_lower : str
        Name of first date column, difference will be calculated as column_upper - column_lower
    column_upper : str
        Name of second date column, difference will be calculated as column_upper - column_lower
    new_column_name : str, default = None
        Name given to calculated datediff column. If None then {column_upper}_{column_lower}_datediff_{units}
        will be used.
    units : str, default = 'D'
        Numpy datetime units, accepted values are 'Y', 'M', 'D', 'h', 'm', 's'
    copy : bool, default = True
        Should X be copied prior to transform?
    verbose: bool, default = False
    """

    def __init__(
        self,
        column_lower,
        column_upper,
        new_column_name=None,
        units="D",
        copy=True,
        verbose=False,
    ):

        if not type(column_lower) is str:

            raise TypeError("column_lower must be a str")

        if not type(column_upper) is str:

            raise TypeError("column_upper must be a str")

        columns = [column_lower, column_upper]

        accepted_values_units = [
            "Y",
            "M",
            "D",
            "h",
            "m",
            "s",
        ]

        if not type(units) is str:

            raise TypeError("units must be a str")

        if units not in accepted_values_units:

            raise ValueError(
                f"units must be one of {accepted_values_units}, got {units}"
            )

        self.units = units

        if new_column_name is not None:

            if not type(new_column_name) is str:

                raise TypeError("new_column_name must be a str")

            self.new_column_name = new_column_name

        else:

            self.new_column_name = f"{column_upper}_{column_lower}_datediff_{units}"

        super().__init__(columns=columns, copy=copy, verbose=verbose)

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper

    def transform(self, X):
        """Calculate the difference between the given fields in the specified units.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        """

        X = super().transform(X)

        X[self.new_column_name] = (
            X[self.columns[1]] - X[self.columns[0]]
        ) / np.timedelta64(1, self.units)

        return X


class ToDatetimeTransformer(BaseTransformer):
    """Class to transform convert specified column to datetime.

    Class simply uses the pd.to_datetime method on the specified column.

    Parameters
    ----------
    column : str
        Name of the column to convert to datetime.

    new_column_name : str
        Name given to the new datetime column.

    to_datetime_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.to_datetime method when it is called in transform.

    **kwargs
        Arbitrary keyword arguments passed onto pd.to_datetime().

    """

    def __init__(self, column, new_column_name, to_datetime_kwargs={}, **kwargs):

        if not type(column) is str:

            raise TypeError(
                "column should be a single str giving the column to transform to datetime"
            )

        if not type(new_column_name) is str:

            raise TypeError("new_column_name must be a str")

        if not type(to_datetime_kwargs) is dict:

            raise TypeError(
                f"to_datetime_kwargs should be a dict but got type {type(to_datetime_kwargs)}"
            )

        else:

            for i, k in enumerate(to_datetime_kwargs.keys()):

                if not type(k) is str:

                    raise TypeError(
                        f"unexpected type ({type(k)}) for to_datetime_kwargs key in position {i}, must be str"
                    )

        self.to_datetime_kwargs = to_datetime_kwargs
        self.new_column_name = new_column_name

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

        super().__init__(columns=[column], **kwargs)

    def transform(self, X):
        """Convert specified column to datetime using pd.to_datetime.

        Parameters
        ----------
        X : pd.DataFrame
            Data with column to transform.

        """

        X = super().transform(X)

        X[self.new_column_name] = pd.to_datetime(
            X[self.columns[0]], **self.to_datetime_kwargs
        )

        return X


class SeriesDtMethodTransformer(BaseTransformer):
    """Tranformer that applies a pandas.Series.dt method.

    Transformer assigns the output of the method to a new column. It is possible to
    supply other key word arguments to the transform method, which will be passed to the
    pandas.Series.dt method being called.

    Be aware it is possible to supply incompatible arguments to init that will only be
    identified when transform is run. This is because there are many combinations of method, input
    and output sizes. Additionally some methods may only work as expected when called in
    transform with specific key word arguments.

    Parameters
    ----------
    new_column_name : str
        The name of the column to be assigned to the output of running the pandas method in transform.

    pd_method_name : str
        The name of the pandas.Series.dt method to call.

    column : str
        Column to apply the transformer to. If a str is passed this is put into a list. Value passed
        in columns is saved in the columns attribute on the object. Note this has no default value so
        the user has to specify the columns when initialising the transformer. This is avoid likely
        when the user forget to set columns, in this case all columns would be picked up when super
        transform runs.

    pd_method_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.Series.dt method when it is called.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    column : str
        Name of column to apply transformer to. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : str
        Column name for transformation.

    new_column_name : str
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    pd_method_name : str
        The name of the pandas.DataFrame method to call.

    pd_method_kwargs : dict
        Dictionary of keyword arguments to call the pd.Series.dt method with.

    """

    def __init__(
        self, new_column_name, pd_method_name, column, pd_method_kwargs={}, **kwargs
    ):

        if type(column) is not str:

            raise TypeError(f"column should be a str but got {type(column)}")

        super().__init__(columns=column, **kwargs)

        if type(new_column_name) is not str:

            raise TypeError(
                f"unexpected type ({type(new_column_name)}) for new_column_name, must be str"
            )

        if type(pd_method_name) is not str:

            raise TypeError(
                f"unexpected type ({type(pd_method_name)}) for pd_method_name, expecting str"
            )

        if type(pd_method_kwargs) is not dict:

            raise TypeError(
                f"pd_method_kwargs should be a dict but got type {type(pd_method_kwargs)}"
            )

        else:

            for i, k in enumerate(pd_method_kwargs.keys()):

                if not type(k) is str:

                    raise TypeError(
                        f"unexpected type ({type(k)}) for pd_method_kwargs key in position {i}, must be str"
                    )

        self.new_column_name = new_column_name
        self.pd_method_name = pd_method_name
        self.pd_method_kwargs = pd_method_kwargs

        try:

            ser = pd.Series([datetime.datetime(2020, 12, 21)])
            getattr(ser.dt, pd_method_name)

        except Exception as err:

            raise AttributeError(
                f"""error accessing "dt.{pd_method_name}" method on pd.Series object - pd_method_name should be a pd.Series.dt method"""
            ) from err

        if callable(getattr(ser.dt, pd_method_name)):

            self._callable = True

        else:

            self._callable = False

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

    def transform(self, X):
        """Transform specific column on input pandas.DataFrame (X) using the given pandas.Series.dt method and
        assign the output back to column in X.

        Any keyword arguments set in the pd_method_kwargs attribute are passed onto the pd.Series.dt method
        when calling it.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column (self.new_column_name) added. These contain the output of
            running the pd.Series.dt method.

        """

        X = super().transform(X)

        if self._callable:

            X[self.new_column_name] = getattr(
                X[self.columns[0]].dt, self.pd_method_name
            )(**self.pd_method_kwargs)

        else:

            X[self.new_column_name] = getattr(
                X[self.columns[0]].dt, self.pd_method_name
            )

        return X


class BetweenDatesTransformer(BaseTransformer):
    """Transformer to generate a boolean column indicating if one date is between two others.

    If not all column_lower values are less than or equal to column_upper when transform is run
    then a warning will be raised.

    Parameters
    ----------
    column_lower : str
        Name of column containing the lower date range values.

    column_between : str
        Name of column to check if it's values fall between column_lower and column_upper.

    column_upper : str
        Name of column containing the upper date range values.

    new_column_name : str
        Name for new column to be added to X.

    lower_inclusive : bool, defualt = True
        If lower_inclusive is True the comparison to column_lower will be column_lower <=
        column_between, otherwise the comparison will be column_lower < column_between.

    upper_inclusive : bool, defualt = True
        If upper_inclusive is True the comparison to column_upper will be column_between <=
        column_upper, otherwise the comparison will be column_between < column_upper.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.__init__().

    Attributes
    ----------
    column_lower : str
        Name of date column to subtract. This attribute is not for use in any method,
        use 'columns' instead. Here only as a fix to allow string representation of transformer.

    column_upper : str
        Name of date column to subtract from. This attribute is not for use in any method,
        use 'columns instead. Here only as a fix to allow string representation of transformer.

    column_between : str
        Name of column to check if it's values fall between column_lower and column_upper. This attribute
        is not for use in any method, use 'columns instead. Here only as a fix to allow string representation of transformer.

    columns : list
        Contains the names of the columns to compare in the order [column_lower, column_between
        column_upper].

    new_column_name : str
        new_column_name argument passed when initialising the transformer.

    lower_inclusive : bool
        lower_inclusive argument passed when initialising the transformer.

    upper_inclusive : bool
        upper_inclusive argument passed when initialising the transformer.

    """

    def __init__(
        self,
        column_lower,
        column_between,
        column_upper,
        new_column_name,
        lower_inclusive=True,
        upper_inclusive=True,
        **kwargs,
    ):

        if type(column_lower) is not str:
            raise TypeError("column_lower should be str")

        if type(column_between) is not str:
            raise TypeError("column_between should be str")

        if type(column_upper) is not str:
            raise TypeError("column_upper should be str")

        if type(new_column_name) is not str:
            raise TypeError("new_column_name should be str")

        if type(lower_inclusive) is not bool:
            raise TypeError("lower_inclusive should be a bool")

        if type(upper_inclusive) is not bool:
            raise TypeError("upper_inclusive should be a bool")

        self.new_column_name = new_column_name
        self.lower_inclusive = lower_inclusive
        self.upper_inclusive = upper_inclusive

        super().__init__(columns=[column_lower, column_between, column_upper], **kwargs)

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column_lower = column_lower
        self.column_upper = column_upper
        self.column_between = column_between

    def transform(self, X):
        """Transform - creates column indicating if middle date is between the other two

        If not all column_lower values are less than or equal to column_upper when transform is run
        then a warning will be raised.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column (self.new_column_name) added. This column is
            boolean and indicates if the middle column is between the other 2.

        """

        X = super().transform(X)

        for col in self.columns:

            if not pd.api.types.is_datetime64_dtype(X[col]):

                raise TypeError(
                    f"{col} should be datetime64[ns] type but got {X[col].dtype}"
                )

        if not (X[self.columns[0]] <= X[self.columns[2]]).all():

            warnings.warn(
                f"not all {self.columns[2]} are greater than or equal to {self.columns[0]}"
            )

        if self.lower_inclusive:

            lower_comparison = X[self.columns[0]] <= X[self.columns[1]]

        else:

            lower_comparison = X[self.columns[0]] < X[self.columns[1]]

        if self.upper_inclusive:

            upper_comparison = X[self.columns[1]] <= X[self.columns[2]]

        else:

            upper_comparison = X[self.columns[1]] < X[self.columns[2]]

        X[self.new_column_name] = lower_comparison & upper_comparison

        return X



"""
This module contains transformers that apply numeric functions.
"""

import numpy as np
import pandas as pd
from sklearn.preprocessing import (
    MinMaxScaler,
    MaxAbsScaler,
    StandardScaler,
    PolynomialFeatures,
)

from tubular.base import BaseTransformer


class LogTransformer(BaseTransformer):
    """Transformer to apply log transformation.

    Transformer has the option to add 1 to the columns to log and drop the
    original columns.

    Parameters
    ----------
    columns : None or str or list
        Columns to log transform.

    base : None or float/int
        Base for log transform. If None uses natural log.

    add_1 : bool
        Should a constant of 1 be added to the columns to be transformed prior to
        applying the log transform?

    drop : bool
        Should the original columns to be transformed be dropped after applying the
        log transform?

    suffix : str, default = '_log'
        The suffix to add onto the end of column names for new columns.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    Attributes
    ----------
    add_1 : bool
        The name of the column or columns to be assigned to the output of running the
        pandas method in transform.

    drop : bool
        The name of the pandas.DataFrame method to call.

    suffix : str
        The suffix to add onto the end of column names for new columns.

    """

    def __init__(
        self, columns, base=None, add_1=False, drop=True, suffix="log", **kwargs
    ):

        super().__init__(columns=columns, **kwargs)

        if base is not None:
            if not isinstance(base, (int, float)):
                raise ValueError("base should be numeric or None")
            if not base > 0:
                raise ValueError("base should be strictly positive")

        self.base = base
        self.add_1 = add_1
        self.drop = drop
        self.suffix = suffix

    def transform(self, X):
        """Applies the log transform to the specified columns.

        If the drop attribute is True then the original columns are dropped. If
        the add_1 attribute is True then the original columns + 1 are logged.

        Parameters
        ----------
        X : pd.DataFrame
            The dataframe to be transformed.

        Returns
        -------
        X : pd.DataFrame
            The dataframe with the specified columns logged, optionally dropping the original
            columns if self.drop is True.

        """

        X = super().transform(X)

        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype, axis=0
        )

        if not numeric_column_types.all():

            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index
            )

            raise TypeError(
                f"The following columns are not numeric in X; {non_numeric_columns}"
            )

        new_column_names = [f"{column}_{self.suffix}" for column in self.columns]

        if self.add_1:

            if (X[self.columns] <= -1).sum().sum() > 0:

                raise ValueError(
                    "values less than or equal to 0 in columns (after adding 1), make greater than 0 before using transform"
                )

            if self.base is None:

                X[new_column_names] = np.log(X[self.columns] + 1)

            else:

                X[new_column_names] = np.log(X[self.columns] + 1) / np.log(self.base)

        else:

            if (X[self.columns] <= 0).sum().sum() > 0:

                raise ValueError(
                    "values less than or equal to 0 in columns, make greater than 0 before using transform"
                )

            if self.base is None:

                X[new_column_names] = np.log(X[self.columns])

            else:

                X[new_column_names] = np.log(X[self.columns]) / np.log(self.base)

        if self.drop:

            X.drop(self.columns, axis=1, inplace=True)

        return X


class CutTransformer(BaseTransformer):
    """Class to bin a column into discrete intervals.

    Class simply uses the [pd.cut](https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.cut.html)
    method on the specified column.

    Parameters
    ----------
    column : str
        Name of the column to discretise.

    new_column_name : str
        Name given to the new discrete column.

    cut_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the pd.cut method when it is called in transform.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init().

    """

    def __init__(self, column, new_column_name, cut_kwargs={}, **kwargs):

        if not type(column) is str:

            raise TypeError(
                "column arg (name of column) should be a single str giving the column to discretise"
            )

        if not type(new_column_name) is str:

            raise TypeError("new_column_name must be a str")

        if not type(cut_kwargs) is dict:

            raise TypeError(
                f"cut_kwargs should be a dict but got type {type(cut_kwargs)}"
            )

        else:

            for i, k in enumerate(cut_kwargs.keys()):

                if not type(k) is str:

                    raise TypeError(
                        f"unexpected type ({type(k)}) for cut_kwargs key in position {i}, must be str"
                    )

        self.cut_kwargs = cut_kwargs
        self.new_column_name = new_column_name

        # This attribute is not for use in any method, use 'columns' instead.
        # Here only as a fix to allow string representation of transformer.
        self.column = column

        super().__init__(columns=[column], **kwargs)

    def transform(self, X):
        """Discretise specified column using pd.cut.

        Parameters
        ----------
        X : pd.DataFrame
            Data with column to transform.

        """

        X = super().transform(X)

        if not pd.api.types.is_numeric_dtype(X[self.columns[0]]):

            raise TypeError(
                f"{self.columns[0]} should be a numeric dtype but got {X[self.columns[0]].dtype}"
            )

        X[self.new_column_name] = pd.cut(X[self.columns[0]], **self.cut_kwargs)

        return X


class ScalingTransformer(BaseTransformer):
    """Transformer to perform scaling of numeric columns.

    Transformer can apply min max scaling, max absolute scaling or standardisation (subtract mean and divide by std).
    The transformer uses the appropriate sklearn.preprocessing scaler.

    Parameters
    ----------
    columns : str, list or None
        Name of the columns to apply scaling to.

    scaler_type : str
        Type of scaler to use, must be one of 'min_max', 'max_abs' or 'standard'. The corresponding
        sklearn.preprocessing scaler used in each case is MinMaxScaler, MaxAbsScaler or StandardScaler.

    scaler_kwargs : dict, default = {}
        A dictionary of keyword arguments to be passed to the scaler object when it is initialised.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init().

    """

    def __init__(self, columns, scaler_type, scaler_kwargs={}, **kwargs):

        if not type(scaler_kwargs) is dict:

            raise TypeError(
                f"scaler_kwargs should be a dict but got type {type(scaler_kwargs)}"
            )

        else:

            for i, k in enumerate(scaler_kwargs.keys()):

                if not type(k) is str:

                    raise TypeError(
                        f"unexpected type ({type(k)}) for scaler_kwargs key in position {i}, must be str"
                    )

        allowed_scaler_values = ["min_max", "max_abs", "standard"]

        if scaler_type not in allowed_scaler_values:

            raise ValueError(f"scaler_type should be one of; {allowed_scaler_values}")

        if scaler_type == "min_max":

            self.scaler = MinMaxScaler(**scaler_kwargs)

        elif scaler_type == "max_abs":

            self.scaler = MaxAbsScaler(**scaler_kwargs)

        elif scaler_type == "standard":

            self.scaler = StandardScaler(**scaler_kwargs)

        # This attribute is not for use in any method
        # Here only as a fix to allow string representation of transformer.
        self.scaler_kwargs = scaler_kwargs
        self.scaler_type = scaler_type

        super().__init__(columns=columns, **kwargs)

    def check_numeric_columns(self, X):
        """Method to check all columns (specicifed in self.columns) in X are all numeric.

        Parameters
        ----------
        X : pd.DataFrame
            Data containing columns to check.

        """

        numeric_column_types = X[self.columns].apply(
            pd.api.types.is_numeric_dtype, axis=0
        )

        if not numeric_column_types.all():

            non_numeric_columns = list(
                numeric_column_types.loc[~numeric_column_types].index
            )

            raise TypeError(
                f"The following columns are not numeric in X; {non_numeric_columns}"
            )

        return X

    def fit(self, X, y=None):
        """Fit scaler to input data.

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe with columns to learn scaling values from.

        y : None
            Required for pipeline.

        """

        super().fit(X, y)

        X = self.check_numeric_columns(X)

        self.scaler.fit(X[self.columns])

        return self

    def transform(self, X):
        """Transform input data X with fitted scaler.

        Parameters
        ----------
        X : pd.DataFrame
            Dataframe containing columns to be scaled.

        Returns
        -------
        X : pd.DataFrame
            Input X with columns scaled.

        """

        X = super().transform(X)

        X = self.check_numeric_columns(X)

        X[self.columns] = self.scaler.transform(X[self.columns])

        return X


class InteractionTransformer(BaseTransformer):
    """Transformer that generates interaction features.
    Transformer generates a new column  for all combinations from the selected columns up to the maximum degree
    provided. (For sklearn version higher than 1.0.0>, only interaction of a degree higher or equal to the minimum
    degree would be computed).
    Each interaction column consists of the product of the specific combination of columns.
    Ex: with 3 columns provided ["a","b","c"], if max degree is 3, the total possible combinations are :
    - of degree 1 : ["a","b","c"]
    - of degree 2 : ["a b","b c","a c"]
    - of degree 3 : ["a b c"]

        Parameters
        ----------
        columns : None or list or str
            Columns to apply the transformer to. If a str is passed this is put into a list. Value passed
            in columns is saved in the columns attribute on the object. Note this has no default value so
            the user has to specify the columns when initialising the transformer. This is avoid likely
            when the user forget to set columns, in this case all columns would be picked up when super
            transform runs.
        min_degree : int
            minimum degree of interaction features to be considered. For example if min_degree=3, only interaction
            columns from at least 3 columns would be generated. NB- only applies if sklearn version is 1.0.0>=
        max_degree : int
            maximum degree of interaction features to be considered. For example if max_degree=3, only interaction
            columns from up to 3 columns would be generated.


         Attributes
        ----------
        min_degree : int
            minimum degree of interaction features to be considered
        max_degree : int
            maximum degree of interaction features to be considered
        nb_features_to_interact : int
            number of selected columns from which interactions should be computed. (=len(columns))
        nb_combinations : int
            number of new interaction features
        interaction_colname : list
            names of each new interaction feature. The name of an interaction feature is the combinations of previous
            column names joined with a whitespace. Interaction feature of ["col1","col2","col3] would be "col1 col2 col3".
        nb_feature_out : int
            number of total columns of transformed dataset, including new interaction features

    """

    def __init__(self, columns, min_degree=2, max_degree=2, **kwargs):

        super().__init__(columns=columns, **kwargs)

        if len(columns) < 2:
            raise ValueError(
                f"number of columns must be equal or greater than 2, got {str(len(columns))} column."
            )

        if type(min_degree) is int:
            if min_degree < 2:
                raise ValueError(
                    f"min_degree must be equal or greater than 2, got {str(min_degree)}"
                )
            else:
                self.min_degree = min_degree
        else:
            raise TypeError(
                f"unexpected type ({type(min_degree)}) for min_degree, must be int"
            )
        if type(max_degree) is int:
            if min_degree > max_degree:
                raise ValueError("max_degree must be equal or greater than min_degree")
            else:
                self.max_degree = max_degree
            if max_degree > len(columns):
                raise ValueError(
                    "max_degree must be equal or lower than number of columns"
                )
            else:
                self.max_degree = max_degree
        else:
            raise TypeError(
                f"unexpected type ({type(max_degree)}) for max_degree, must be int"
            )

        self.nb_features_to_interact = len(self.columns)
        self.nb_combinations = -1
        self.interaction_colname = []
        self.nb_feature_out = -1

    def transform(self, X):
        """Generate from input pandas DataFrame (X) new interaction features using the "product" pandas.DataFrame method
         and add this column or columns in X.

        Parameters
        ----------
        X : pd.DataFrame
            Data to transform.

        Returns
        -------
        X : pd.DataFrame
            Input X with additional column or columns (self.interaction_colname) added. These contain the output of
            running the  product pandas DataFrame method on identified combinations.

        """
        X = super().transform(X)

        try:
            interaction_combination_index = PolynomialFeatures._combinations(
                n_features=self.nb_features_to_interact,
                min_degree=self.min_degree,
                max_degree=self.max_degree,
                interaction_only=True,
                include_bias=False,
            )
        except TypeError as err:
            if (
                str(err)
                == "_combinations() got an unexpected keyword argument 'min_degree'"
            ):
                interaction_combination_index = PolynomialFeatures._combinations(
                    n_features=self.nb_features_to_interact,
                    degree=self.max_degree,
                    interaction_only=True,
                    include_bias=False,
                )
            else:
                raise err

        interaction_combination_colname = [
            [self.columns[col_idx] for col_idx in interaction_combination]
            for interaction_combination in interaction_combination_index
        ]
        self.nb_combinations = len(interaction_combination_colname)
        self.nb_feature_out = self.nb_combinations + len(X)

        self.interaction_colname = [
            " ".join(interaction_combination)
            for interaction_combination in interaction_combination_colname
        ]

        for inter_idx in range(len(interaction_combination_colname)):
            X[self.interaction_colname[inter_idx]] = X[
                interaction_combination_colname[inter_idx]
            ].product(axis=1, skipna=False)

        return X



from tubular.base import BaseTransformer


class SetValueTransformer(BaseTransformer):
    """Transformer to set value of column(s) to a given value.

    This should be used if columns need to be set to a constant value.

    Parameters
    ----------
    columns: list or str
        Columns to set values.

    value : various
        Value to set.

    **kwargs
        Arbitrary keyword arguments passed onto BaseTransformer.init method.

    """

    def __init__(self, columns, value, **kwargs):

        self.value = value

        super().__init__(columns=columns, **kwargs)

    def transform(self, X):
        """Set columns to value.

        Parameters
        ----------
        X : pd.DataFrame
            Data to apply mappings to.

        Returns
        -------
        X : pd.DataFrame
            Transformed input X with columns set to value.

        """

        X = super().transform(X)

        X[self.columns] = self.value

        return X
